/**
 *  Created:     10. 11. 2018 by Petra Sečkařová (xsecka02@stud.fit.vutbr.cz).
 *  Filename:    formula.cpp
 *
 *  Description: Implementation of methods of a class representing a single ptLTL formula.
 */

#include "formula.h"

/**
 * A constructor setting this formula's id and initializing it's state and operation to ATOM.
 *
 * @param id The UNIQUE id to be set to this formula.
 */
formula::formula(int id) : id(id) {
    operation = ATOM;
    s = CREATED;
}

/**
 * Adds new subformula id to the list of operands, updates initialization state accordingly.
 *
 * @param value An id of subformula, that is an operand to this one.
 */
void formula::addOperand(int value) {

    //! Update initialization state
    switch (operation) {

        // unary operations
        case LAST:
        case PREVIOUSLY:
        case GLOBALLY:
        case NOT:
        case UP:
        case DOWN:
            s = READY;
            break;

        // binary operations
        case S_SINCE:
        case W_SINCE:
        case S_INTERVAL:
        case W_INTERVAL:
        case AND:
        case OR:
        case XOR:
        case IMPLY:
            if (operands.empty()) s = FIRST_OP;
            else s = READY;
            break;

        case ATOM:break; // shouldn't ever happen - always asking if canAddOperands() before calling this
    }

    //! Add the operand
    operands.push_back(value);
}

/**
 * Constructs a code for one step of processing of this formula in a created monitor.
 * The returned string is based on recursive expressions for each supported operation
 * introduced in http://www.havelund.com/Publications/sttt-tacas02.pdf
 *
 * @return An expression constructed for recursive evaluation of this formula.
 */
std::string formula::getContent() {
    std::stringstream ss; /*< string stream for the returned expression construction */
    std::string sep; /*< a separator string for cyclic writing of operators by operations that can have 2+ of them */
    switch (operation) {
        case LAST:
            ss << "M.pre[" << operands[0] << "]";
            content = ss.str();
            break;
        case PREVIOUSLY:
            ss << "M.now[" << operands[0] << "]" << " || M.pre[" << id << "]";
            content = ss.str();
            break;
        case GLOBALLY:
            ss << "M.now[" << operands[0] << "]" << " && M.pre[" << id << "]";
            content = ss.str();
            break;
        case S_SINCE:
        case W_SINCE:
            ss << "M.now[" << operands[1] << "]" << " || (M.now[" << operands[0] << "] && M.pre[" << id << "])";
            content = ss.str();
            break;
        case S_INTERVAL:
        case W_INTERVAL:
            ss << "!M.now[" << operands[1] << "]" << " && (M.pre[" << id << "] || M.now[" << operands[0] << "])" ;
            content = ss.str();
            break;
        case UP:
            ss << "!M.pre[" << operands[0] << "] && M.now[" << operands[0] << "]";
            content = ss.str();
            break;
        case DOWN:
            ss << "M.pre[" << operands[0] << "] && !M.now[" << operands[0] << "]";
            content = ss.str();
            break;
        case AND:
        case OR:
        case XOR:
            for (auto operand : operands)
            {
                ss << sep << "M.now[" << operand << "]";
                if (operation == AND) sep = " && ";
                else if (operation == OR) sep = " || ";
                else  sep = " ^ ";
            }
            content = ss.str();
            break;
        case NOT:
            ss << "!M.now[" << operands[0] << "]";
            content = ss.str();
            break;
        case IMPLY:
            ss << "!M.now[" << operands[0] << "] || M.now[" << operands[1] << "]";
            content = ss.str();
            break;
        case ATOM:break; // is set already
    }
    return content;
}


/**
 * Checks, that the number of operands is correct or content not empty if operation == ATOM.
 *
 * @return True if nothing is missing, false otherwise.
 */
bool formula::isFullySet() {
    if (operation == ATOM) return !content.empty();
    else return s == READY;
}


/**
 * Checks if any other operands can be added.
 *
 * @return True if there can be any more operands added, false otherwise.
 */
bool formula::canAddOperands() {
    return operation != ATOM && (s != READY || operation == OR || operation == AND || operation == XOR);
}


/**
 * A setter of this formula's content. Meant for ATOM formulae, won't have any useful effect for others.
 * The passed name of state variable is given a prefix _e_ used in enum created for generated monitor
 * and wrapped in M.values[...] which denotes a reference to function returning its value computed
 * as defined in specification file.
 *
 * @param c An atomic expression, typically a name of a boolean variable to be set as this formula's content.
 */
void formula::setContent(std::string c) {
    if(!c.empty()) {
        if (c[0] == '!')
            content = "!M.values[_e_" + c.substr(1,c.size()-1) + "]()";
        else
            content = "M.values[_e_" + c + "]()";

        print = c;
    }

}


/**
 * Constructs a string for initialization of processing of this formula in a created monitor.
 * The returned string is based on recursive expressions for each supported operation
 * introduced in http://www.havelund.com/Publications/sttt-tacas02.pdf
 *
 * @return An expression constructed for initialization of evaluation of this formula.
 */
std::string formula::getInitContent() {
    std::stringstream ss; /*< string stream for the returned expression construction */
    std::string sep; /*< a separator string for cyclic writing of operators by operations that can have 2+ of them */
    switch (operation) {
        case LAST:
        case PREVIOUSLY:
        case GLOBALLY:
            ss << "M.pre[" << operands[0] << "]";
            content = ss.str();
            break;
        case S_SINCE:
            ss << "M.pre[" << operands[1] << "]";
            content = ss.str();
            break;
        case W_SINCE:
            ss << "M.pre[" << operands[1] << "]" << " || M.pre[" << operands[0] << "]";
            content = ss.str();
            break;
        case S_INTERVAL:
            ss << "!M.pre[" << operands[1] << "]" << " && M.pre[" << operands[0] << "] " ;
            content = ss.str();
            break;
        case W_INTERVAL:
            ss << "!M.pre[" << operands[1] << "]";
            content = ss.str();
            break;
        case UP:
        case DOWN:
            ss << "false";
            content = ss.str();
            break;
        case AND:
        case OR:
        case XOR:
            for (auto operand : operands)
            {
                ss << sep << "M.pre[" << operand << "]";
                if (operation == AND) sep = " && ";
                else if (operation == OR) sep = " || ";
                else  sep = " ^ ";
            }
            content = ss.str();
            break;
        case NOT:
            ss << "!M.pre[" << operands[0] << "]";
            content = ss.str();
            break;
        case IMPLY:
            ss << "!M.pre[" << operands[0] << "] || M.pre[" << operands[1] << "]";
            content = ss.str();
            break;
        case ATOM:break; // Content prepared by default
    }
    return content;
}


/**
 * A decoder of OpType enum into a string representation.
 *
 * @return A string representation of this formula's operation ended by one space if not ATOM.
 */
std::string formula::printOpType() {
    switch (operation) {

        case LAST:
            return "L ";
        case PREVIOUSLY:
            return "P ";
        case GLOBALLY:
            return "G ";
        case S_SINCE:
            return "Ss ";
        case W_SINCE:
            return "Sw ";
        case UP:
            return "U ";
        case DOWN:
            return "D ";
        case S_INTERVAL:
            return "Is ";
        case W_INTERVAL:
            return "Iw ";
        case AND:
            return "&& ";
        case OR:
            return "|| ";
        case XOR:
            return "^ ";
        case NOT:
            return "! ";
        case IMPLY:
            return "-> ";
        default:
            return print;
    }
}

/**
 * A method incrementing all ids by 1.
 * Used when after successfully parsing one formula, another one appears in a specification file,
 * which is not connected to the previous one, so a new root formula is created. The root formula
 * has to have id = 0, so all formulae parsed until now must increment their ids.
 */
void formula::incrementIds() {
    // Increment id of this formula
    ++id;

    // Increment ids of all operands
    for (auto &id : operands) {
        ++id;
    }
}

State formula::getState() {
    return s;
}

int formula::getId() {
    return id;
}

void formula::setPrint(std::string p) {
    print = std::move(p);
}

std::vector<int> &formula::getOperands() {
    return operands;
}

std::string formula::getPrintText() {
    return print;
}

OpType formula::getOperation() {
    return operation;
}

void formula::setOperation(OpType op) {
    s = INITIALIZED;
    operation = op;
}

const fileLocation &formula::getStart() const {
    return start;
}

void formula::setStart(const fileLocation &start) {
    formula::start = start;
}

const fileLocation &formula::getEnd() const {
    return end;
}

void formula::setEnd(const fileLocation &end) {
    formula::end = end;
}


