/**
 *  Created:     10. 11. 2018 by Petra Sečkařová (xsecka02@stud.fit.vutbr.cz).
 *  Filename:    formula.cpp
 *
 *  Description: Declaration of class representing a single ptLTL formula and needed enums.
 */

#ifndef DP_FORMULA_H
#define DP_FORMULA_H


#include <vector>
#include <string>
#include <sstream>
#include <utility>
#include "fileLocation.h"

/**
 * An enum encoding past-time LTL and common logic operators.
 * 'S_' means strong, 'W_' is weak.
 */
enum OpType {
    LAST,
    PREVIOUSLY,
    GLOBALLY,
    S_SINCE,
    W_SINCE,
    UP,
    DOWN,
    S_INTERVAL,
    W_INTERVAL,
    AND,
    OR,
    XOR,
    NOT,
    IMPLY,
    ATOM /*< Boolean expression - should be a name of a variable */
};

/**
 * An enum used by formulaAnalyser for decisions in formulae processing.
 */
enum State {
    CREATED,     /*< After constructor */
    INITIALIZED, /*< Operation was set, doesn't say anything about operands */
    FIRST_OP,    /*< One operator is set, doesn't say anything about the operation */
    READY        /*< Everything is set - operation and corresponding number of operands */
};

/**
 * A class representing one ptLTL formula.
 * The formula can be defined either
 *        * as one operator and a list of operands
 *        * as a state variable defined in specification or passed .h file
 * Each formula has a unique numeric id, provides it's string representation,
 * starting and ending file location in the specification file it was read from
 * and a state of initialization for formulaAnalyzer.
 */
class formula {

private:
    int id; /*< A UNIQUE id of formula - every it's subformula has a greater id value
             *< Is always set in constructor
             *< The main (root) formula has id == 0 */

    OpType operation; /*< Operator of this formula defined using enum OpType */

    std::vector<int> operands; /*< A vector of direct-subformulae ids
                                *< is empty for ATOM, permits 2+ operands for AND, OR and XOR */

    std::string content; /*< A string representation of the formula */

    std::string print;   /*< A string used by ATOM formulae to store unmodified
                          *< name of the represented variable */

    State s; /*< A state of initialization of the formula using enum State */

    fileLocation start; /*< A location of first character of this formula in the specification file */

    fileLocation end; /*< A location of last character of this formula in the specification file */

public:

    /**
     * A constructor setting this formula's id and initializing it's state and operation to ATOM.
     *
     * @param id The UNIQUE id to be set to this formula.
     */
    explicit formula(int id);

    /**
     * Adds new subformula id to the list of operands, updates initialization state accordingly.
     *
     * @param value An id of subformula, that is an operand to this one.
     */
    void addOperand(int);

    /**
     * Constructs a code for one step of processing of this formula in a created monitor.
     * The returned string is based on recursive expressions for each supported operation
     * introduced in http://www.havelund.com/Publications/sttt-tacas02.pdf
     *
     * @return An expression constructed for recursive evaluation of this formula.
     */
    std::string getContent();

    /**
     * Constructs a string for initialization of processing of this formula in a created monitor.
     * The returned string is based on recursive expressions for each supported operation
     * introduced in http://www.havelund.com/Publications/sttt-tacas02.pdf
     *
     * @return An expression constructed for initialization of evaluation of this formula.
     */
    std::string getInitContent();

    /**
     * Checks, that the number of operands is correct or content not empty if operation == ATOM.
     *
     * @return True if nothing is missing, false otherwise.
     */
    bool isFullySet();

    /**
     * Checks if any other operands can be added.
     *
     * @return True if there can be any more operands added, false otherwise.
     */
    bool canAddOperands();

    /**
     * A decoder of OpType enum into a string representation.
     *
     * @return A string representation of this formula's operation ended by one space if not ATOM.
     */
    std::string printOpType();


    /**
     * A method incrementing all ids by 1.
     * Used when after successfully parsing one formula, another one appears in a specification file,
     * which is not connected to the previous one, so a new root formula is created. The root formula
     * has to have id = 0, so all formulae parsed until now must increment their ids.
     */
    void incrementIds();

    /*************************************** GET ***************************************/

    int getId();

    State getState();

    const fileLocation &getStart() const;

    const fileLocation &getEnd() const;

    OpType getOperation();

    std::vector<int> &getOperands();

    std::string getPrintText();

    /*************************************** SET ***************************************/


    void setOperation(OpType);


    /**
     * A setter of this formula's content. Meant for ATOM formulae, won't have any useful effect for others.
     * The passed name of state variable is given a prefix _e_ used in enum created for generated monitor
     * and wrapped in M.values[...] which denotes a reference to function returning its value computed
     * as defined in specification file.
     *
     * @param c An atomic expression, typically a name of a boolean variable to be set as this formula's content.
     */
    void setContent(std::string);


    void setPrint(std::string);


    void setStart(const fileLocation &start);


    void setEnd(const fileLocation &end);


    /**
     * An < operator definition based on id's.
     * Note that a formula is considered to be less if it's id is !GREATER!.
     * This is used for sorting the formulae by id in DESCENDING order.
     *
     * @param other The other formula to be compared.
     * @return True if this formula's id is greater than the other's one.
     */
    bool operator< (const formula &other) const {
        return id > other.id;
    }

    /**
     * An == operator definition based on id's.
     *
     * @param other The other formula to be compared.
     * @return True if this formula's id equal to the other's one.
     */
    bool operator== (const formula &other) const {
        return id == other.id;
    }
};


#endif //DP_FORMULA_H