/**
 *  Created:     8. 10. 2018 by Petra Sečkařová (xsecka02@stud.fit.vutbr.cz).
 *  Filename:    formulaAnalyser.h
 *
 *  Description: Implementation of methods of a class for parsing and analysing the specification file.
 */

#include "formulaAnalyser.h"

/**
 * Parses data from given file. Loads all formulae and variable definitions into internal
 * structures of this class for later usage.
 *
 * @param filename A name of file where from the data should be parsed.
 * @return True if all data was successfully read and properly stored, false otherwise.
 */
bool formulaAnalyser::parseFormula(std::string filename) {

    std::ifstream formulaFile (filename);
    if (formulaFile.is_open())
    {
        char c; /*< for reading the formulae from data file char by char */
        std::string word;
        std::string content;
        std::string line;
        int nextId = 0;
        bool isAtom = false;
        formula *currentFormula = nullptr;
        int lineNo = 1;
        int column = 0;
        int missingParentheses = 0;

        std::stack<formula> formulaStack;
        std::vector<int> activeFormulae;
        bool definitions = false;
        std::regex defsRgx(R"(\s*(\S+)\s*:\s*(.*))"); // parsing definitions
        std::regex srcRgx(R"(\s*[Ss][Oo][Uu][Rr][Cc][Ee][Ss]\s*:\s*)"); // sources annotation line
        std::regex defsAnnotationRgx(R"(\s*[Dd][Ee][Ff][Ii][Nn][Ii][Tt][Ii][Oo][Nn][Ss]\s*:\s*)"); // defs annotation line
        std::smatch matches;
        bool sourcesListPresent = false;


        std::size_t found = filename.find_last_of("/\\");
        std::string path = found == std::string::npos ? "" : filename.substr(0,found+1);


        //! Parse the formulae
        while (!definitions && formulaFile.get(c))
        {
            column++;
            switch(c)
            {
                //! Start of a new formula
                case '(':
                    if (!word.empty())
                    { // no whitespace between the '(' and previous word
                        OpType op = decodeType(word);
                        if (op == ATOM)
                        {
                            std::cout << "Invalid operator: " << word << std::endl;
                            return false;
                        } else if (op == NOT && (currentFormula == nullptr || currentFormula->getState() != CREATED)) {
                            addNewFormula(formulaStack, activeFormulae, nextId, currentFormula);
                            currentFormula->setStart(fileLocation(lineNo,column-1,filename));
                            missingParentheses++;
                        }
                        currentFormula->setOperation(op);
                    }

                    content.clear();
                    word.clear();

                    if(!addNewFormula(formulaStack, activeFormulae, nextId, currentFormula)) {
                        std::cout << "Too many operands for formula (" << currentFormula->printOpType() << "... starting on "
                        << currentFormula->getStart().printLocation() << "! " <<  std::endl;
                        return false;
                    }
                    currentFormula->setStart(fileLocation(lineNo,column,filename));
                    break;

                // ~ case '('


                //! End of current formula
                case ')':

                    if (!word.empty())
                    // no space between last operand and ')' -> add the operand
                    {
                        if (currentFormula->canAddOperands())
                        { // last operand == variable
                            generateAtom(word, nextId, currentFormula, formulaStack, activeFormulae,
                                         lineNo, column, filename);

                        }
                        else if (currentFormula->getOperation() == ATOM)
                        { // cannot add operands - this formula is a variable or !variable
                            if (!currentFormula->isFullySet()) {
                                currentFormula->setContent(content + word);
                                noteVar(word);
                            }
                            else {
                                std::cout << "Invalid operator: " << content << std::endl;
                                return false;
                            }
                        }
                        else
                        { // Trying to add too many operands to currentFormula
                            std::cout << "Too many operands for formula (" << currentFormula->printOpType() << "... starting on "
                                      << currentFormula->getStart().printLocation() << "! " <<  std::endl;
                            return false;
                        }
                    }

                    // no active formula - nothing to be closed
                    if (formulaStack.empty()) {
                        std::cout << "Too many ')' at the end of line " << lineNo << std::endl;
                        return false;
                    }

                    //! Close current formula
                    currentFormula->setEnd(fileLocation(lineNo,column,filename));

                    if(!finalizeCurrentFormula(currentFormula, &formulaStack, &activeFormulae))
                    {
                        std::cout << "Incomplete formula " << currentFormula->printOpType() << " starting on "
                        << currentFormula->getStart().printLocation() << "!" << std::endl;
                        return false;
                    }
                    // if finalizing currentFormula succeeded, reassign the pointer to the new top of formulaStack
                    else if (!formulaStack.empty()) currentFormula = &formulaStack.top();


                    if(currentFormula->getOperation() == NOT && missingParentheses > 0) {
                        missingParentheses--;
                        currentFormula->setEnd(fileLocation(lineNo,column,filename));
                        finalizeCurrentFormula(currentFormula, &formulaStack, &activeFormulae);
                        if (!formulaStack.empty()) currentFormula = &formulaStack.top();

                    }

                    word.clear();
                    content.clear();
                    break;

                // ~ case ')'

                case '\n':
                case '\t':
                case '\r':
                case ' ': // might be end of current word

                    if (c == '\n') {
                        lineNo++;
                        column = 0;
                    }

                    if (word.empty()) continue; // more white characters in sequence
                    else if(std::regex_match(word, defsAnnotationRgx) || std::regex_match(word, srcRgx)) {
                        // end of formulae part of the input file

                        sourcesListPresent = std::regex_match(word, srcRgx);
                        definitions = true; //! Start loading variable definitions
                        continue;
                    }

                    if (formulaStack.empty()) {
                        addNewFormula(formulaStack, activeFormulae, nextId, currentFormula);
                    }

                    if (currentFormula->getState() == CREATED) // first word: operator or id
                    {
                        OpType op = decodeType(word);
                        isAtom = op == ATOM;
                        currentFormula->setOperation(op);
                    }
                    else if (currentFormula->getOperation() == NOT)
                    { // ( ! id )
                        currentFormula->setOperation(ATOM);
                        isAtom = true;
                        content = "!";
                    }
                    else if (currentFormula->getOperation() != ATOM) { // operand ends w/ white space -> id
                        generateAtom(word, nextId, currentFormula, formulaStack, activeFormulae, lineNo, column,
                                     filename);
                    }
                    else {
                        // current operation == ATOM, more words than one!
                        std::cout << "Invalid operator: " << content << std::endl;
                        return false;
                    }

                    if (isAtom) {
                        content += word;
                        currentFormula->setContent(content);
                        noteVar(word);
                    }
                    word.clear();
                    break;

                // ~ case whitespaces

                default:
                    word += c;
                    break;

            }
        }

        //! check that all formulae are parsed and stored
        if (!finalizeFormulaeParsing(currentFormula, formulaStack, activeFormulae))
            return false;

        //! Load variable definitions
        int idx;
        while (std::getline(formulaFile, line)) {
            if(!line.empty()) {
                if (!sourcesListPresent) {
                    if (std::regex_match(line, srcRgx))
                        sourcesListPresent = true;
                    else {
                        if(!std::regex_search(line, matches, defsRgx)) {
                            std::cout << "Invalid variable definition: " << line << std::endl;
                            return false; // necessary?
                        }

                        idx = getIndexOfVarName(std::string("_e_").append(matches[1]));
                        if (idx < 0)
                            std::cout << "Found a definition of a variable that does not "
                                         "appear in the formulae: " << matches[0] << std::endl;
                        else {
                            varDefinitions[idx] = matches[2].str();
                        }
                    }
                }
                else {
                    line.erase(0, line.find_first_not_of(" \t\r\v\f"));
                    srcList.push_back(path + line);
                }

            }
        }

        formulaFile.close();

        //! Store full string version of each formula
        std::sort(formulae.begin(), formulae.end());
        std::string print;
        for (auto &f : formulae) {
            print = "(" + f.printOpType();
            std::string sep;
            for (auto op : f.getOperands()) {
                print += sep + formulae[formulae.size() - 1 - op].getPrintText();
                sep = " ";
            }
            print += ")";
            f.setPrint(print);
            if (f.getStart().getLine() >= 0) {
                std::string pos;
                pos = std::to_string(f.getStart().getLine()) + ":" +
                          std::to_string(f.getStart().getColumn()) + "-" +
                          std::to_string(f.getEnd().getColumn()) + ":";

                std::string fpos = f.getStart().getFileName();
                fpos.append(":");
                fpos += pos;
                fpos.append("\\n\\t");
                fpos += print;
                print = fpos;

                print += R"(\n\tturned %s after being %s for last %d states\n)";
            }
            contents.emplace_back(print);
        }
    }
    else {
        std::cout << "Unable to open given file with formula!" << std::endl;
        return false;
    }

    return true;
}

/**
 * Checks that all formulae are successfully parsed and no unfinished is left.
 * There can be a single formula on the stack, but it has to be in the READY state.
 *
 * @param currentFormula A formula at the top of the formulaStack if there is any left.
 * @param formulaStack A stack of formulae to be processed.
 * @param activeFormulae A list of ids of all formulae in formulaStack.
 * @return True if the formulaStack was emptied correctly, false otherwise.
 */
bool formulaAnalyser::finalizeFormulaeParsing(formula *currentFormula, std::stack<formula> &formulaStack,
                                              std::vector<int> &activeFormulae) {
    if (formulaStack.size() > 1)
    {
        std::cout << "Incomplete formula " << currentFormula->printOpType() << " starting on "
               << currentFormula->getStart().printLocation() << "!" << std::endl;

        return false;
    }
    else if (!formulaStack.empty())
    {
        if((currentFormula->getOperation() != ATOM && !addedRoot) || !finalizeCurrentFormula(currentFormula, &formulaStack, &activeFormulae))
        {
            std::cout << "Incomplete formula (" << currentFormula->printOpType() << "... starting on "
                      << currentFormula->getStart().printLocation() << "!" << std::endl;
            return false;
        }
    }
    return true;
}

/**
 * Generates an atomic formula from given string that is considered to be a name of a variable.
 * The new formula is added as an operand to the formula currently on the top of the formulaStack.
 * Given string is noted as a variable name along with mapping it's appearance in currently active formulae.
 *
 * @param word A content of an atomic formula to be added by this function.
 * @param nextId A counter used for assigning ids to the formulae.
 * @param currentFormula A formula at the top of the formulaStack. The newly created formula will be added
 * as this formula's operand.
 * @param formulaStack A stack of formulae currently being processed.
 * @param activeFormulae A list of ids of all formulae in formulaStack.
 */
bool formulaAnalyser::generateAtom(std::string &word, int &nextId, formula *&currentFormula,
                                   std::stack<formula> &formulaStack, std::vector<int> &activeFormulae, int line,
                                   int column, std::string &filename) {

    if (!addNewFormula(formulaStack, activeFormulae, nextId, currentFormula)) return false;
    currentFormula->setOperation(ATOM);
    currentFormula->setContent(word);
    currentFormula->setStart(fileLocation(line, int(column-word.length()), filename));
    currentFormula->setEnd(fileLocation(line, column-1, filename));
    noteVar(word);
    // no need for checking success of next line (we've just set everything needed for it to succeed)
    finalizeCurrentFormula(currentFormula, &formulaStack, &activeFormulae);
    currentFormula = &formulaStack.top();
    return true;
}

/**
 * If the current formula is fully set, than this method stores it into the list of parsed formulae,
 * pops it from the stack and erases it from the list of active formulae.
 *
 * @param currentFormula A formula at the top of the formulaStack to be stored.
 * @param formulaStack A stack of not-yet-ready formulae where from the currentFormula should be removed.
 * @param activeFormulae A list of ids of all formulae in the formulaStack.
 * @return True if currentFormula was successfully stored, false otherwise.
 */
bool formulaAnalyser::finalizeCurrentFormula(formula *currentFormula, std::stack<formula> *formulaStack,
                                             std::vector<int> *activeFormulae)  {
    if (currentFormula->isFullySet())
    {
        auto it = std::find(activeFormulae->begin(), activeFormulae->end(),currentFormula->getId());
        if (it != activeFormulae->end())
            activeFormulae->erase(it);
        formulae.push_back(*currentFormula);
        int childrenCnt = (int)currentFormula->getOperands().size();
        if (childrenCnt > maxChildCnt) maxChildCnt = childrenCnt;
        formulaStack->pop();
        return true;
    }
    else return false;
}

/**
 * Pushes a new formula onto the given stack, adds it's id to the activeFormulae list
 * and updates the operands of previous top of the stack accordingly. If other formulae were parsed
 * before already, but the formulaStack is empty, adds a new root formula with the operator &&
 * and correctly updates its operands and all previous formulae' ids.
 *
 * @param formulaStack A stack of formulae currently being processed.
 * @param activeFormulae A list of ids of the formulae, that are currently on the formulaStack.
 * @param nextId A counter used for assigning ids of the parsed formulae, is incremented by this funcion.
 * @param currentFormula A pointer to a formula currently at the top of formulaStack.
 */
bool formulaAnalyser::addNewFormula(std::stack<formula> &formulaStack, std::vector<int> &activeFormulae, int &nextId,
                                    formula *&currentFormula) {
    if (!formulaStack.empty()) {
        if (!currentFormula->canAddOperands()) return false;
        currentFormula->addOperand(nextId);
    }
    else if (!formulae.empty()) {
        //! More separate formulae given in specification
        // Create one root formula that performs && with the results of all separate formulae.

        // activeFormulae and formulaStack are empty
        activeFormulae.push_back(0); // the new root formula
        formulaStack.push(formula(0));
        currentFormula = &formulaStack.top();
        addedRoot = true;

        // Increment ids and id references in all formulae that had already been processed
        for (auto &f : formulae) {
            f.incrementIds();
        }
        // set operation of current root as && of previous root (previous id 0) and nextId+1
        currentFormula->setOperation(AND);
        currentFormula->addOperand(1);
        // added the new root formula, current nextId is already used (incremented ids)
        currentFormula->addOperand(++nextId);
    }

    activeFormulae.push_back(nextId);
    formulaStack.push(formula(nextId++));
    currentFormula = &formulaStack.top();
    return true;
}

/**
 * Decodes string representation of operator in a formula into an OpType enum value.
 *
 * @param op A string representation of the operator.
 * @return An enum value corresponding to the given string representation of operator.
 */
OpType formulaAnalyser::decodeType(const std::string &op) {
    if (op == "G")
        return GLOBALLY;
    else if (op == "P")
        return PREVIOUSLY;
    else if (op == "L")
        return LAST;
    else if (op == "Ss")
        return S_SINCE;
    else if (op == "Sw")
        return W_SINCE;
    else if (op == "Is")
        return S_INTERVAL;
    else if (op == "Iw")
        return W_INTERVAL;
    else if (op == "U")
        return UP;
    else if (op == "D")
        return DOWN;
    else if (op == "||")
        return OR;
    else if (op == "!")
        return NOT;
    else if (op == "&&")
        return AND;
    else if (op == "^")
        return XOR;
    else if (op == "->")
        return IMPLY;
    else return ATOM;
}

/**
 * Ensures, that given name of a variable is stored. Updates map of appearance of given
 * variable in formulae.
 *
 * @param name A name of variable encountered in some part of a formula.
 * @param activeFormulae A list of ids of current parent formulae.
 */
void formulaAnalyser::noteVar(std::string name) {
    if (name[0] == '!') // '!id'
        name = name.substr(1,name.size()-1);

    // in case of using actual variable of the SUT, the enum value-name can't be the same
    std::string enumName = "_e_" + name;

    if (varNames.emplace(enumName).second) {
        varDefinitions.push_back(name);
    }
}

/**
 * Computes the index of a variable in varNames (which is the same as for varDefinitions)
 * given by it's name.
 *
 * @param name A name of the variable, which index is being searched.
 * @return The index of a variable with given name in varNames vector. If no such name
 * found, the value -1 is returned.
 */
int formulaAnalyser::getIndexOfVarName(std::string name) {
    auto it = varNames.find(name);
    if(it != varNames.end())
        return static_cast<int>(std::distance(varNames.begin(), it));
    return -1;
}

std::vector<formula> formulaAnalyser::getFormulae() {
    return formulae;
}

std::set<std::string> formulaAnalyser::getVarNames() {
    return varNames;
}

unsigned formulaAnalyser::getFormulaeCnt() {
    return formulae.size();
}

std::vector<std::string> formulaAnalyser::getFormulaeContents() {
    return contents;
}

std::vector<std::string> formulaAnalyser::getVarDefinitions() {
    return varDefinitions;
}

std::vector<std::string> formulaAnalyser::getSrcList() {
    return srcList;
}

int formulaAnalyser::getMaxChildCnt() const {
    return maxChildCnt;
}
