/**
 *  Created:     8. 10. 2018 by Petra Sečkařová (xsecka02@stud.fit.vutbr.cz).
 *  Filename:    formulaAnalyser.h
 *
 *  Description: Declaration of a class for parsing and analysing the specification file.
 */

#ifndef DP_FORMULAANALYSER_H
#define DP_FORMULAANALYSER_H

#include <string>
#include <set>
#include <map>
#include <vector>
#include <stack>
#include <climits>
#include <fstream>
#include <iostream>
#include <algorithm>
#include <regex>
#include "formula.h"


/**
 * Class for parsing files with formulae into internal structures. Stores the parsed formulae
 * in an implicit tree (given by order and index references) of subformulae, string representation
 * of each formula and names of all variable names encountered in the formulae.
 */
class formulaAnalyser {
private:
    std::set<std::string> varNames; /*< All names of variables appearing in parsed formulae */

    std::vector<std::string> varDefinitions;

    std::vector<std::string> srcList;

    std::vector<formula> formulae; /*< All parsed formulae divided into subformulae.
                                    *< Together building a tree given by ids in the 'operands' list*/

    std::vector<std::string> contents; /*< String representation of each formula from previous list */

    int maxChildCnt;
    bool addedRoot;

public:
    /**
     * Parses data from given file. Loads all formulae and variable definitions into internal
     * structures of this class for later usage.
     *
     * @param filename A name of file where from the data should be parsed.
     * @return True if all data was successfully read and properly stored, false otherwise.
     */
    bool parseFormula(std::string filename);

    /**
     * Checks that all formulae are successfully parsed and no unfinished is left.
     * There can be a single formula on the stack, but it has to be in the READY state.
     *
     * @param currentFormula A formula at the top of the formulaStack if there is any left.
     * @param formulaStack A stack of formulae to be processed.
     * @param activeFormulae A list of ids of all formulae in formulaStack.
     * @return True if the formulaStack was emptied correctly, false otherwise.
     */
    bool finalizeFormulaeParsing(formula *currentFormula, std::stack<formula> &formulaStack,
                                 std::vector<int> &activeFormulae);

    /**
     * Generates an atomic formula from given string that is considered to be a name of a variable.
     * The new formula is added as an operand to the formula currently on the top of the formulaStack.
     * Given string is noted as a variable name along with mapping it's appearance in currently active formulae.
     *
     * @param word A content of an atomic formula to be added by this function.
     * @param nextId A counter used for assigning ids to the formulae.
     * @param currentFormula A formula at the top of the formulaStack. The newly created formula will be added
     * as this formula's operand.
     * @param formulaStack A stack of formulae currently being processed.
     * @param activeFormulae A list of ids of all formulae in formulaStack.
     */
    bool generateAtom(std::string &word, int &nextId, formula *&currentFormula,
                      std::stack<formula> &formulaStack, std::vector<int> &activeFormulae, int line,
                      int column, std::string &filename);

    /**
     * If the current formula is fully set, than this method stores it into the list of parsed formulae,
     * pops it from the stack and erases it from the list of active formulae.
     *
     * @param currentFormula A formula at the top of the formulaStack to be stored.
     * @param formulaStack A stack of not-yet-ready formulae where from the currentFormula should be removed.
     * @param activeFormulae A list of ids of all formulae in the formulaStack.
     * @return True if currentFormula was successfully stored, false otherwise.
     */
    bool finalizeCurrentFormula(formula *currentFormula, std::stack<formula> *formulaStack,
                                std::vector<int> *activeFormulae);

    /**
     * Pushes a new formula onto the given stack, adds it's id to the activeFormulae list
     * and updates the operands of previous top of the stack accordingly. If other formulae were parsed
     * before already, but the formulaStack is empty, adds a new root formula with the operator &&
     * and correctly updates its operands and all previous formulae' ids.
     *
     * @param formulaStack A stack of formulae currently being processed.
     * @param activeFormulae A list of ids of the formulae, that are currently on the formulaStack.
     * @param nextId A counter used for assigning ids of the parsed formulae, is incremented by this funcion.
     * @param currentFormula A pointer to a formula currently at the top of formulaStack.
     */
    bool addNewFormula(std::stack<formula> &formulaStack, std::vector<int> &activeFormulae, int &nextId,
                       formula *&currentFormula);


    /**
     * Decodes string representation of operator in a formula into an OpType enum value.
     *
     * @param op A string representation of the operator.
     * @return An enum value corresponding to the given string representation of operator.
     */
    OpType decodeType(const std::string &op);

    /**
     * Ensures, that given name of a variable is stored. Updates map of appearance of given
     * variable in formulae.
     *
     * @param name A name of variable encountered in some part of a formula.
     */
    void noteVar(std::string name);

    /**
     * Computes the index of a variable in varNames (which is the same as for varDefinitions)
     * given by it's name.
     *
     * @param name A name of the variable, which index is being searched.
     * @return The index of a variable with given name in varNames vector. If no such name
     * found, the value -1 is returned.
     */
    int getIndexOfVarName(std::string name);

    std::vector<formula> getFormulae();

    std::set<std::string> getVarNames();

    unsigned getFormulaeCnt();

    std::vector<std::string> getVarDefinitions();

    std::vector<std::string> getFormulaeContents();


    int getMaxChildCnt() const;

    std::vector<std::string> getSrcList();
};


#endif //DP_FORMULAANALYSER_H
