/**
 *  Created:     24. 9. 2018 by Petra Sečkařová (xsecka02@stud.fit.vutbr.cz).
 *  Filename:    formulaAnalyser.h
 *
 *  Description: Implementation of methods of a class for generating the code of monitor.
 */

#include "monitorCreator.h"
/**
 * Retrieves a size in bytes of a file given by it's name.
 *
 * @param filename A name of file which size should be returned.
 * @return A size of the file in bytes.
 */
long fileSize(const char *filename) {
    std::ifstream in(filename, std::ifstream::ate | std::ifstream::binary);
    return in.tellg();
}


/**
 * A constructor initializing the formula filename.
 */
monitorCreator::monitorCreator(std::string f) {
    formulaFile = std::move(f);
}


/**
 * Creates a C-language program consisting of two source files: monitor.h
 * and monitor.c. The created header contains declaration of functions named
 * monitorVerify() and reset(). Function monitorVerify() should be called
 * by the SUT on every change of it's state to trigger the verification.
 * The function reset() can be used by the SUT to erase the monitor's
 * memory of errors. The file monitor.c defines monitor structure according
 * to the input formulae, initializes it and also provides definition of
 * the monitorVerify() and reset() functions along with other needed functions.
 *
 * @param defsFile A path to header file containing declarations of all variables
 * and functions used given formulae.
 * @return True if monitor files generated successfully, false otherwise.
 */
bool monitorCreator::printMonitor(const std::string &defsFile) {

    //! Parse formulae
    formulaAnalyser analyser = formulaAnalyser();

    if (!analyser.parseFormula(formulaFile)) return false;

    //! Replace instrumentation comments in given list of source files
    replaceInstrumentationComments(analyser.getSrcList());

    //! Generate the monitor implementation
    std::string comment = getCommentForGeneratedFiles(defsFile);

    generateHeaderFile(defsFile, comment);

    generateMonitorImplementation(analyser, comment);

    return true;
}

/**
 * Generates a header file with the passed comment, #include "defsFile" and declaration
 * of the functions monitorVerify() and reset().
 *
 * @param defsFile A path to header file containing declarations of all variables
 * and functions used given formulae.
 * @param comment A comment generated by getCommentForGeneratedFiles.
 */
void monitorCreator::generateHeaderFile(const std::string &defsFile, const std::string &comment) const {//! Print monitor header file
    FILE* headerFile = fopen(GENERATED_FILES_NAME ".h","w");

    //! See monitorCreator.h:112-123
    // makes only the function 'monitorVerify()' accessible
    fprintf(headerFile, HEADER_FILE, comment.c_str(), defsFile.c_str());

    fclose(headerFile);
}

/**
 * Generates the monitor.c file with definitions of needed structures and functions
 * for verification of the specified temporal properties.
 *
 * @param analyser A formulaAnalyser that has successfully parsed a specification file.
 * @param comment A comment generated by getCommentForGeneratedFiles.
 */
void monitorCreator::generateMonitorImplementation(formulaAnalyser &analyser, const std::string &comment) {//! Print the implementation
    FILE* monitor = fopen(GENERATED_FILES_NAME ".c","w");

    fprintf(monitor, MONITOR_C_HEADER, comment.c_str());

    //! Define functions for evaluation
    int i = 0;
    for (const auto &def : analyser.getVarDefinitions()) {
        std::fprintf(monitor, "bool f%d(){ return %s; }\n", i, def.c_str());
        i++;
    }

    int formulaeCnt = (int)analyser.getFormulaeCnt();
    printVarNamesEnum(analyser.getVarNames(), monitor);

    //! Print the monitor structure
    printMonitorStructure(monitor, analyser.getFormulae(), analyser.getFormulaeContents(),
                          analyser.getVarDefinitions(),
                          analyser.getMaxChildCnt());


    //! Print fail reporting method
    std::fprintf(monitor, FAIL_REPORT_METHOD, formulaeCnt-1);

    //! Print method encoding formulae values preparation
    printPreparation(monitor, analyser.getFormulae());
    
    //! Print method encoding formulae one evaluation step
    printProcessing(monitor, analyser.getFormulae());

    //! Print reset method
    std::fprintf(monitor, RESET_METHOD);

    std::fprintf(monitor, "#ifdef __cplusplus\n}\n#endif\n\n");

    fclose(monitor);
}


/**
 * Prints enumeration of variable names used in evaluated formulae for easier
 * access to their values while evaluating.
 *
 * @param varNames A formulaAnalyser object filled with parsed formulae, etc.
 * @param monitor A stream pointing to the file to write into.
 */
void monitorCreator::printVarNamesEnum(std::set<std::string> varNames, FILE *monitor) {
    std::string sep;
    std::ostringstream os;
    for (auto &name : varNames) {
        os << sep << name;
        sep = ", ";
    }
    std::fprintf(monitor,ENUM_VAR_NAMES,os.str().c_str());
}

/**
 * Prints definition of a monitor structure into given file. First defines
 * functions for evaluation of the variables' values, then prints
 * the structure, creates it's instance M and initializes all it's needed values.
 *
 * @param monitor A file to write into.
 * @param formulae The vector of monitored formulae.
 * @param contents Strings representing the contents of monitored formulae.
 * @param varsDefinitions Code segments defining how to evaluate used variables
 * in order given by enumeration created by printVarNamesEnum().
 */
void monitorCreator::printMonitorStructure(FILE *monitor, std::vector<formula> formulae,
                                           std::vector<std::string> contents, std::vector<std::string> varsDefinitions,
                                           int childrenCnt) {

    int formulaeCnt = (int)formulae.size();

    //! Declare the monitor struct
    std::fprintf(monitor,MONITOR_STRUCT_DECLARATION,(int)varsDefinitions.size(),
            formulaeCnt, formulaeCnt, formulaeCnt, formulaeCnt, formulaeCnt, formulaeCnt,
            childrenCnt,formulaeCnt);

    //! Define the monitor struct values

    //! Get list of state variables' values
    std::ostringstream values;
    std::string sep = "\n\t\t";
    for (int i = 0; i < (int)varsDefinitions.size(); i++) {
        values << sep << "&f" << i;
        sep = ",\n\t\t";

    }

    //! Get list of formulae contents strings
    std::ostringstream contentsList;
    sep = "\n\t\t";
    std::reverse(contents.begin(),contents.end());
    for (auto &c : contents) {
        contentsList << sep << "\"" << c <<  "\\n\"";
        sep = ",\n\t\t";
    }

    //! Get lists of formulae children
    std::ostringstream children;
    sep = "";
    for (auto it = formulae.rbegin(); it != formulae.rend(); ++it ) {
        children << sep << "\n\t\t{";
        sep = "";
        for (auto ch : (*it).getOperands()) {
            children << sep << ch;
            sep = ", ";
        }
        if ((*it).getOperands().empty()) children << "0";
        children << "}";
        sep = ", ";
    }

    //! Get list of formulae children counts
    std::ostringstream childrenCounts;
    sep = "";
    for (auto it = formulae.rbegin(); it != formulae.rend(); ++it ) {
        childrenCounts << sep << (*it).getOperands().size();
        sep = ", ";
    }

    //! Print the monitor structure definition
    std::fprintf(monitor,MONITOR_STRUCT_DEFINITION,values.str().c_str(),
            contentsList.str().c_str(), children.str().c_str(), childrenCounts.str().c_str());
}


/**
 * Prints method initiating the boolean values of all formulae from 'previous step'
 * for the evaluation in the first round of evaluation.
 *
 * @param monitor A file to write into.
 * @param formulae A list of formulae sorted so, that the subformula always comes
 * before the formula that contains it.
 */
void monitorCreator::printPreparation(FILE *monitor, std::vector<formula> formulae) {

    std::fprintf(monitor,"void prepare() {\n\n");

    //! Perform initialization of the pre formula states vector values
    for (auto &f : formulae)
    {
        int id = f.getId();
        fprintf(monitor, "    M.pre[%d] = %s;\n", id, f.getInitContent().c_str());
    }

    //! Set flag that prev values are prepared
    std::fprintf(monitor,"    initiated = true;\n}\n\n");
}


/**
 * Prints method performing processing of evaluation of the formulae
 * encoded in the created program using values from the last step of evaluation.
 *
 * @param monitor A file to write into.
 * @param formulae A list of formulae sorted so, that the subformula always comes
 * before the formula that contains it.
 */
void monitorCreator::printProcessing(FILE *monitor, std::vector<formula> formulae) {

    //! Update the variables' values and swap now and pre vector values
    //! if already initialized, else call prepare()
    std::fprintf(monitor, MONITOR_VERIFY_INIT, (int)formulae.size());

    //! Perform one step of evaluation for each formula
    for (auto &f : formulae)
    {
        int id = f.getId();
        std::fprintf(monitor, "    M.now[%d] = %s;\n", id, f.getContent().c_str());
    }

    //! Update since vector values and check if state valid
    std::fprintf(monitor, MONITOR_VERIFY_CHECK_AND_FINALIZE, (int)formulae.size());
}


/**
 * Creates a comment header for generated files informing how and when were
 * these files generated.
 *
 * @param defsFile The name of given definitions header file.
 * @return The string containing created comment
 */
std::string monitorCreator::getCommentForGeneratedFiles(const std::string &defsFile) {

    time_t now = time(nullptr);

    return "/**\n * Generated by monitorCreator class of T-Props Checker.\n"
           " * Command: ./spectra --tpc -s " + formulaFile + " -d " + defsFile + "\n * Date: " +
           ctime(&now) + " */\n\n";
}

/**
 * Replaces all occurrences of '// T-Props Checker verify' (case insensitive)
 * in given file with calls of function monitorVerify() of the generated monitor.
 *
 * @param filename A name of file where these labels should be replaced.
 */
void monitorCreator::replaceLabelsInFileWithMonitorCalls(const std::string &filename) {

    std::regex label(R"(//\s*[Tt]\s*-\s*[Pp]rops\s+[Cc]hecker\s+[Vv]erify\s*)");
    std::regex cpp(R"(.*\.cpp)");

    std::ofstream replaced("tmp");
    std::ifstream src(filename);

    std::string line, last_include;
    bool monitor_header_included = false;

    //! Copy given file with replaced labels into temporary file named 'tmp'
    if (src.is_open())
        while (std::getline(src, line)) {
            if (line.find("#include ") != std::string::npos) {
                if (line.find("monitor.h\"") != std::string::npos)
                    monitor_header_included = true;
                else
                    last_include = line;
            }

            replaced << std::regex_replace(line, label, "monitorVerify();\n") << std::endl;
        }

    src.close();
    replaced.close();

    //! Insert the #include "monitor.h" line if not included before
    if (!monitor_header_included) {

        rename("tmp", "tmp1");
        std::ofstream completed("tmp");
        std::ifstream old("tmp1");

        if (old.is_open())
            while (std::getline(old, line)) {

                completed << line << std::endl;
                if (line == last_include) {
                    //! wrap in extern "C" { ... } if the source is a C++ code
                    //! to make the compiler understand the compiled monitor code
                    bool is_cpp = std::regex_match(filename,cpp);
                    if (is_cpp)
                        completed << "extern \"C\" {" << std::endl;
                    completed << "#include \"monitor.h\"" << std::endl;
                    if (is_cpp)
                        completed << "}" << std::endl;

                    std::cout << "An inclusion of monitor.h was added into "  << filename <<
                    ". Please modify the path to this header in this source correspondingly "
                    "or place the header in the same folder as this source." << std::endl;
                }
            }

        old.close();
        completed.close();
        remove("tmp1");

    }

    //! If any changes made, save previous version with '.old' extension and rewrite
    if (fileSize(filename.c_str()) != fileSize("tmp")) {
        std::string newName = filename+".old";
        rename(filename.c_str(), newName.c_str());
        rename("tmp",filename.c_str());
    }
    else { // No changes made -> remove temporary file
        remove("tmp");
    }

}

/**
 * Calls replaceLabelsInFileWithMonitorCalls for each file on srcList.
 *
 * @param srcList A list of source files paths where the instrumentation
 * comments can be placed and should be replaced.
 */
void monitorCreator::replaceInstrumentationComments(const std::vector<std::string> &srcList) {
    if (!srcList.empty()) {
        for (auto &file : srcList) {
            replaceLabelsInFileWithMonitorCalls(file);
        }
    }
}


void monitorCreator::setFormulaFile(const std::string &formulaFile) {
    monitorCreator::formulaFile = formulaFile;
}

