/**
 *  Created:     24. 9. 2018 by Petra Sečkařová (xsecka02@stud.fit.vutbr.cz).
 *  Filename:    formulaAnalyser.h
 *
 *  Description: Declaration of a class for generating the code of monitor.
 */

#ifndef DP_MONITORCREATOR_H
#define DP_MONITORCREATOR_H

#include <string>
#include <fstream>
#include <utility>
#include <stack>
#include <regex>
#include <cstdio>
#include "formulaAnalyser.h"

#define GENERATED_FILES_NAME "monitor"

/**
 * Retrieves a size in bytes of a file given by it's name.
 *
 * @param filename A name of file which size should be returned.
 * @return A size of the file in bytes.
 */
long fileSize(const char *filename);

/**
 * A class constructing a code of monitor for run-time evaluation
 * of specified temporal properties.
 */
class monitorCreator {
private:
    std::string formulaFile; /*< A name of specification file containing the formulae to be monitored */

public:
    /**
     * A constructor initializing the formula filename attribute.
     */
    explicit monitorCreator(std::string);

    /**
     * Creates a C-language program consisting of two source files: monitor.h
     * and monitor.c. The created header contains declaration of functions named
     * monitorVerify() and reset(). Function monitorVerify() should be called
     * by the SUT on every change of it's state to trigger the verification.
     * The function reset() can be used by the SUT to erase the monitor's
     * memory of errors. The file monitor.c defines monitor structure according
     * to the input formulae, initializes it and also provides definition of
     * the monitorVerify() and reset() functions along with other needed functions.
     *
     * @param defsFile A path to header file containing declarations of all variables
     * and functions used given formulae.
     * @return True if monitor files generated successfully, false otherwise.
     */
    bool printMonitor(const std::string &defsFile);

    /**
     * Generates a header file with the passed comment, #include "defsFile" and declaration
     * of the functions monitorVerify() and reset().
     *
     * @param defsFile A path to header file containing declarations of all variables
     * and functions used given formulae.
     * @param comment A comment generated by getCommentForGeneratedFiles.
     */
    void generateHeaderFile(const std::string &defsFile, const std::string &comment) const;

    /**
     * Generates the monitor.c file with definitions of needed structures and functions
     * for verification of the specified temporal properties.
     *
     * @param analyser A formulaAnalyser that has successfully parsed a specification file.
     * @param comment A comment generated by getCommentForGeneratedFiles.
     */
    void generateMonitorImplementation(formulaAnalyser &analyser, const std::string &comment);

    /**
     * Prints enumeration of variable names used in evaluated formulae for easier
     * access to their values while evaluating.
     *
     * @param varNames A formulaAnalyser object filled with parsed formulae, etc.
     * @param monitor A stream pointing to the file to write into.
     */
    void printVarNamesEnum(std::set<std::string> varNames, FILE *monitor);

    /**
     * Prints definition of a monitor structure into given file. First defines
     * functions for evaluation of the variables' values, then prints
     * the structure, creates it's instance M and initializes all it's needed values.
     *
     * @param monitor A file to write into.
     * @param formulae The vector of monitored formulae.
     * @param contents Strings representing the contents of monitored formulae.
     * @param varsDefinitions Code segments defining how to evaluate used variables
     * in order given by enumeration created by printVarNamesEnum().
     */
    void printMonitorStructure(FILE *monitor, std::vector<formula> formulae,
                               std::vector<std::string> contents, std::vector<std::string> varsDefinitions,
                               int childrenCnt);

    /**
     * Prints method initiating the boolean values of all formulae from 'previous step'
     * for the evaluation in the first round of evaluation.
     *
     * @param monitor A file to write into.
     * @param formulae A list of formulae sorted so, that the subformula always comes
     * before the formula that contains it.
     */
    void printPreparation(FILE *monitor, std::vector<formula> formulae);

    /**
     * Prints method performing processing of evaluation of the formulae
     * encoded in the created program using values from the last step of evaluation.
     *
     * @param monitor A file to write into.
     * @param formulae A list of formulae sorted so, that the subformula always comes
     * before the formula that contains it.
     */
    void printProcessing(FILE *monitor, std::vector<formula> formulae);


    /**
     * Creates a comment header for generated files informing how and when were
     * these files generated.
     *
     * @param defsFile The name of given definitions header file.
     * @return The string containing created comment
     */
    std::string getCommentForGeneratedFiles(const std::string &defsFile);

    /**
     * Replaces all occurrences of '// T-Props Checker verify' (case insensitive)
     * in given file with calls of function monitorVerify() of the generated monitor.
     *
     * @param filename A name of file where these labels should be replaced.
     */
    void replaceLabelsInFileWithMonitorCalls(const std::string &filename);

    /**
     * Calls replaceLabelsInFileWithMonitorCalls for each file on srcList.
     *
     * @param srcList A list of source files paths where the instrumentation
     * comments can be placed and should be replaced.
     */
    void replaceInstrumentationComments(const std::vector<std::string> &srcList);

    void setFormulaFile(const std::string &formulaFile);

};

/*********** Templates for generating code of the monitor **********/

#define HEADER_FILE "%s\n" /* comment - when and how was this generated */\
                "#ifndef MONITOR_H\n"\
                "#define MONITOR_H\n\n"\
                "#include <stdio.h>\n"\
                "#include <string.h>\n"\
                "#include <stdbool.h>\n"\
                "#include \"%s\"\n\n" /* name of given header file with definitions */ \
                "#ifdef __cplusplus\n"\
                "extern \"C\" {\n#endif\n\n"\
                "   bool monitorVerify();\n\n"\
                "#ifdef __cplusplus\n}\n#endif\n\n"\
                "#endif"

#define MONITOR_C_HEADER "%s\n" /* comment - when and how was this generated */\
"#include \"" GENERATED_FILES_NAME ".h\"\n\n"\
"#ifdef __cplusplus\n"\
"extern \"C\" {\n"\
"#endif\n\n"

#define ENUM_VAR_NAMES "\ntypedef bool (*fce)();\n"\
"\n"\
"enum VarNames { %s };\n\n"

#define MONITOR_STRUCT_DECLARATION "typedef struct {\n"\
"    fce values[%d];\n" /* count of state variables */ \
"    const char* formulaeContents[%d];\n" /* count of formulae */\
"    bool pre[%d], now[%d], check[%d];\n" /* 3x count of formulae */\
"    int since[%d];\n" /* count of formulae */\
"    int children[%d][%d];\n"/* count of formulae, max. count of children */\
"    int chCnt[%d];\n" /* count of formulae */\
"    int hb;\n"\
"} monitor;\n\n"

#define MONITOR_STRUCT_DEFINITION "monitor M = {\n"\
"    {%s\n" /* values of state variables */\
"    },\n" \
"    {%s\n" /* strings representing contents of formulae */\
"    },\n" \
"    {false},\n" /* pre */\
"    {false},\n" /* now */\
"    {false},\n" /* check */\
"    {0},\n"     /* since */\
"    {%s\n"/* lists of children */\
"    },\n" \
"    { %s },\n"/* counts of children */\
"    0\n" /* initial heartbeat number */\
"};\n\n" \
"bool initiated = false;\n\n"

#define MONITOR_VERIFY_INIT "bool monitorVerify() {\n" \
"    if (!initiated) {prepare();}\n" \
"    else {\n" \
"        for (int i = 0; i < %d; i++) {\n" /* count of formulae */\
"            if (M.pre[i] != M.now[i])\n"\
"                M.since[i] = M.hb;\n\n"\
"            M.pre[i] = M.now[i]; \n" \
"        }\n" \
"        M.hb++;\n" \
"    }\n\n" \
"\n"

#define MONITOR_VERIFY_CHECK_AND_FINALIZE "\n" \
"    for (int i = 0; i < %d; i++) {\n" /* count of formulae */\
"        if ((M.check[i] || i == 0) && M.now[i] != M.pre[i]) {\n" \
"            M.check[i] = true;\n" \
"            for (int j = 0; j < M.chCnt[i]; j++) {\n" \
"                M.check[M.children[i][j]] = true;\n" \
"            }\n" \
"        } else M.check[i] = false;\n" \
"    }\n" \
"\n" \
"    if (!M.now[0]) printFailReport();\n" \
"    return M.now[0];\n" \
"}\n"

#define FAIL_REPORT_METHOD "\n"\
"void printFailReport() {\n"\
"    fprintf(stderr,\"Verification failed after round #%%d!\\nRelevant changes:\\n\", M.hb);\n"\
"\n"\
"    int i;\n"\
"    /*< Print all changes in formulae values that happened in the failure heartbeat */\n"\
"    for (i = %d; i > 0 ; i--) {\n" /* count of formulae - 1 -> the root is not printed*/\
"        if (M.check[i]) {\n"\
"            fprintf(stderr, M.formulaeContents[i], (M.now[i]?\"true\":\"false\"), (M.pre[i]?\"true\":\"false\"),"\
" M.hb - M.since[i]); \n"\
"        }\n"\
"    }\n"\
"}\n\n"

#define RESET_METHOD "\n"\
"void reset() {\n"\
"    initiated = false;"\
"}\n\n"

#endif //DP_MONITORCREATOR_H
