#include "spectra.h"

/**
 * === SPECTRA ===
 * A set of tools for automated translation of specification to a structured data.
 * Runs a tool as selected by given program arguments.
 *
 * Implemented tools:
 *      --tpc : T-Props Checker - Checks specified temporal properties.
 *
 * @param argc The count of program arguments.
 * @param argv argv The contents of the given arguments.
 * @return 1 on any error, 0 otherwise.
 */
int main(int argc, char** argv) {

    if (argc < 2) {
        // -> no arguments
        printHelp();
        return 0;

    } else if (strcmp(argv[1],"--tpc") == 0) {

        //! Temporal Properties Checker

        std::string fileNames[2] = {FORMULA_FILE, DEFS_FILE};

        if (argc == 2) {
            printTPCHelp();
        } else if (parsedArgsForTPC(argc,argv,fileNames)) { // override fileNames if given in program

            //! Generate the monitor structure
            monitorCreator mCreator = monitorCreator(fileNames[FORMULAE]);

            //! Print monitor files
            if (!mCreator.printMonitor(fileNames[DECLARATIONS])) return 1;

            //! Create static library
            std::string cmd = "cc -Wall -fPIC -c " GENERATED_FILES_NAME
                    ".c; ar cr libspectra.a " GENERATED_FILES_NAME ".o;echo rm -f "
                    GENERATED_FILES_NAME  ".o " GENERATED_FILES_NAME ".c";

            // Debug:
            std::cerr << "Building monitor:\n" << cmd << std::endl;

            // Execute compiler
            return system(cmd.c_str());
        } else {
            return 1; // error message already printed
        }
    }

    //! HERE insert calls to other implemented tools
    // check the arguments, print some more detailed help if that is the case
    // if arguments set correctly perform the tool's work

    return 0;
}


/**
 * Prints help message to stdout.
 */
void printHelp() {
    std::cout << " === SPECTRA ===\n"
                 "A set of tools for automated translation of specification to a structured data.\n"
                 "\n"
                 "Implemented tools:\n"
                 " * T-Props Checker: --tpc\n"
                 "          A tool for checking temporal properties of finite runs of programs\n"
                 "\n"
                 "For detailed usage description please see README or try the desired switch\n"
                 "without any other arguments." << std::endl;
}

/**
 * Prints help message for T-Props Checker to stdout.
 */
void printTPCHelp() {
    std::cout << " === SPECTRA : T-Props Checker ===\n"
                 "A tool for checking temporal properties of finite runs of programs\n"
                 "\n"
                 "Usage:\n"
                 "\n"
                 " ./spectra --tpc -i path/to/specification -d path/to/definitions\n"
                 "\n"
                 " where\n"
                 "\n"
                 " -i path/to/specification\n"
                 "      provides a path to a file, where the temporal properties to be checked\n"
                 "      are specified in the format given by \n"
                 "              spectra/doc/T-Props_Checker/InputFileSpec.md\n"
                 "\n"
                 " -d path/to/definitions\n"
                 "      provides a path to header file containing declarations of any variables\n"
                 "      or functions used for specifying state variables of the checked system.\n"
                 "      That means any variable or function used as R-value in the second part \n"
                 "          (see the spectra/doc/T-Props_Checker/InputFileSpec.md file!)\n"
                 "      of the specification or in the formulae section without being further\n"
                 "      defined in the second part.\n"
                 "\n"
                 "Examples of usage and inclusion to a project can be found in spectra/test/" << std::endl;
}

/**
 * Tries to parse arguments for T-Props Checker tool.
 * Loads names of definitions header and formulae file given as '-d defs.h' and '-f formulae.ltl' in any order.
 * @param argc The count of given arguments.
 * @param argv The contents of the given arguments, argv[0] == name of this program, argv[1] == '--tpc'.
 * @param fileNames A pointer where to store the loaded file names.
 * @return True if arguments processed successfully, false otherwise.
 */
bool parsedArgsForTPC(int argc, char** argv, std::string* fileNames) {
    if (argc == 6) { // if file names given

        int opt;
        bool declarationsSet= false, formulaeSet = false;

        //! Load the arguments' values
        while ((opt = getopt(5, &argv[1], "s:d:")) != -1)
        {
            switch (opt)
            {
                case 's':
                    fileNames[FORMULAE] = std::string(optarg);
                    formulaeSet = true;
                    break;
                case 'd':
                    fileNames[DECLARATIONS] = std::string(optarg);
                    declarationsSet = true;
                    break;
                case '?':
                    if (optopt == 's' || optopt == 'd')
                        fprintf (stderr, "Option -%c requires an argument.\n", optopt);
                    else if (isprint (optopt))
                        fprintf (stderr, "Unknown option `-%c'.\n", optopt);
                    else
                        fprintf (stderr,
                                 "Unknown option character `\\x%x'.\n",
                                 optopt);
                    return false;
                default:
                    return false;
            }
        }
        if (!formulaeSet || !declarationsSet) {
            std::cerr << "You must provide both specification and definitions file!\n"
                         "Please see README or run './spectra --tpc' for a detailed usage description." << std::endl;
            return false;
        }

    } else {
        std::cerr << "Wrong count of arguments for T-Props Checker!\n"
                     "For a detailed usage description please see README or try the --tpc switch\n"
                     "without any other arguments." << std::endl;
        return false;
    }

    return true;
}
