//
// Created by seckarovap on 4/2/19.
//

#include "formulaAnalyserTest.h"

#define ID 42

formulaAnalyserTest::formulaAnalyserTest() {
    analyser = formulaAnalyser();
}


/**
 * Function noteVar stores all given names as varNames.
 */
TEST_F(formulaAnalyserTest,NoteVar_notes_all_names) {
    ASSERT_TRUE(analyser.getVarNames().empty());

    std::set<std::string> varNames = {"name1", "name2", "name3"};
    std::set<std::string> expectedOutput = {"_e_name1", "_e_name2", "_e_name3"};

    for (const auto &name : varNames) {
        analyser.noteVar(name);
    }

    ASSERT_EQ(expectedOutput,analyser.getVarNames());
}

/**
 * Function noteVar stores every given name just once.
 */
TEST_F(formulaAnalyserTest,NoteVar_notes_each_name_only_once) {
    ASSERT_TRUE(analyser.getVarNames().empty());
    std::set<std::string> varNames = {"name1", "name2", "name3"};
    std::set<std::string> expectedOutput = {"_e_name1", "_e_name2", "_e_name3"};

    for (const auto &name : varNames) {
        analyser.noteVar(name);
    }

    for (const auto &name : varNames) {
        analyser.noteVar(name);
    }

    ASSERT_EQ(expectedOutput,analyser.getVarNames());
}

TEST_F(formulaAnalyserTest, AddNewFormula_as_first) {

    std::stack<formula> formulaStack;
    std::vector<int> activeFormulae;
    formula *currentFormula = nullptr;
    int nextId = ID;


    ASSERT_TRUE(formulaStack.empty());
    ASSERT_TRUE(activeFormulae.empty());

    analyser.addNewFormula(formulaStack, activeFormulae, nextId, currentFormula);

    ASSERT_FALSE(formulaStack.empty());
    ASSERT_FALSE(activeFormulae.empty());
    ASSERT_EQ(currentFormula,&formulaStack.top());
    ASSERT_EQ(currentFormula->getId(),ID);
    ASSERT_EQ(nextId, ID+1);
    ASSERT_NE(std::find(activeFormulae.begin(), activeFormulae.end(), ID), activeFormulae.end());
}


TEST_F(formulaAnalyserTest, AddNewFormula_as_Nth) {

    formula root = formula(0);
    root.setOperation(AND);
    root.addOperand(1);

    std::stack<formula> formulaStack;
    formulaStack.push(root);
    std::vector<int> activeFormulae = {0};
    formula *currentFormula = &formulaStack.top();
    int nextId = ID;

    analyser.addNewFormula(formulaStack, activeFormulae, nextId, currentFormula);

    // current formula is correctly assigned
    ASSERT_EQ(currentFormula,&formulaStack.top());
    ASSERT_EQ(currentFormula->getId(),ID);

    currentFormula->setContent("var");
    analyser.finalizeCurrentFormula(currentFormula, &formulaStack, &activeFormulae);
    currentFormula = &formulaStack.top();

    // new formula added as operand to last active formula
    ASSERT_NE(std::find(currentFormula->getOperands().begin(), currentFormula->getOperands().end(), ID), root.getOperands().end());
}


TEST_F(formulaAnalyserTest, AddNewFormula_not_first_but_stack_empty) {

    // create some already parsed and closed formula
    formula root = formula(0);
    root.setOperation(ATOM);
    root.setContent("var1");

    std::stack<formula> formulaStack;
    formulaStack.push(root);
    std::vector<int> activeFormulae = {0};
    formula *currentFormula = &root;

    analyser.finalizeCurrentFormula(currentFormula, &formulaStack, &activeFormulae);

    // formula with id 0 saved, none other are 'active'
    ASSERT_EQ(analyser.getFormulae()[0], root);
    ASSERT_TRUE(activeFormulae.empty());
    ASSERT_TRUE(formulaStack.empty());

    int nextId = ID;

    // create new formula on the empty stack
    analyser.addNewFormula(formulaStack, activeFormulae, nextId, currentFormula);

    // check the old root's id - new root is added, all ids are incremented by 1
    ASSERT_EQ(analyser.getFormulae()[0].getId(), 1);

    // finalize the current formula and check the new root
    currentFormula->setOperation(ATOM);
    currentFormula->setContent("var2");
    analyser.finalizeCurrentFormula(currentFormula, &formulaStack, &activeFormulae);

    ASSERT_FALSE(formulaStack.empty());

    currentFormula = &formulaStack.top();

    // new formula and old root added as operands to the new root
    ASSERT_NE(std::find(currentFormula->getOperands().begin(), currentFormula->getOperands().end(), ID+1),
              currentFormula->getOperands().end());
    ASSERT_NE(std::find(currentFormula->getOperands().begin(), currentFormula->getOperands().end(), 1),
              currentFormula->getOperands().end());
}


TEST_F(formulaAnalyserTest, FinalizeCurrentFormula_not_fully_set) {

    std::stack<formula> formulaStack;
    std::vector<int> activeFormulae;
    formula *currentFormula = nullptr;
    int nextId = ID;

    analyser.addNewFormula(formulaStack, activeFormulae, nextId, currentFormula);
    currentFormula->setOperation(AND);

    ASSERT_FALSE(analyser.finalizeCurrentFormula(currentFormula,&formulaStack, &activeFormulae));

    currentFormula->addOperand(0);

    ASSERT_FALSE(analyser.finalizeCurrentFormula(currentFormula,&formulaStack, &activeFormulae));

}


TEST_F(formulaAnalyserTest, FinalizeCurrentFormula_fully_set) {

    std::stack<formula> formulaStack;
    std::vector<int> activeFormulae;
    formula *currentFormula = nullptr;
    int nextId = ID;

    std::string name = "var";

    analyser.addNewFormula(formulaStack, activeFormulae, nextId, currentFormula);
    currentFormula->setOperation(ATOM);
    currentFormula->setContent("var");

    ASSERT_NE(std::find(activeFormulae.begin(), activeFormulae.end(), ID),
              activeFormulae.end());

    ASSERT_TRUE(analyser.finalizeCurrentFormula(currentFormula,&formulaStack, &activeFormulae));

    ASSERT_EQ(std::find(activeFormulae.begin(), activeFormulae.end(), ID),
              activeFormulae.end());

    ASSERT_TRUE(formulaStack.empty());
    ASSERT_FALSE(analyser.getFormulae().empty());
    ASSERT_EQ(analyser.getFormulae()[0].getId(),ID);

    formula reference = formula(ID);
    reference.setOperation(ATOM);
    reference.setContent(name);

    ASSERT_EQ(analyser.getFormulae()[0].getContent(),reference.getContent());

}


TEST_F(formulaAnalyserTest, GenerateAtom_stack_not_empty) {

    formula root = formula(0);
    root.setOperation(AND);

    std::stack<formula> formulaStack;
    formulaStack.push(root);
    std::vector<int> activeFormulae = {0};
    formula *currentFormula = &formulaStack.top();
    int nextId = ID;

    std::string name = "var";
    std::string filename = "file";

    analyser.generateAtom(name, nextId, currentFormula, formulaStack, activeFormulae, 1, 1, filename);

    ASSERT_EQ(currentFormula->getOperation(), AND);
    ASSERT_EQ(currentFormula->getOperands().size(), 1);
    ASSERT_FALSE(analyser.getFormulae().empty());
    ASSERT_EQ(currentFormula->getOperands()[0],ID);

    formula reference = formula(ID);
    reference.setOperation(ATOM);
    reference.setContent(name);

    ASSERT_EQ(analyser.getFormulae()[0].getContent(),reference.getContent());

}


TEST_F(formulaAnalyserTest, GenerateAtom_empty_stack) {

    std::stack<formula> formulaStack;
    std::vector<int> activeFormulae;
    formula *currentFormula = nullptr;
    int nextId = ID;

    std::string name = "var";
    std::string filename = "file";

    analyser.generateAtom(name, nextId, currentFormula, formulaStack, activeFormulae,
                          1, 1, filename);

    ASSERT_TRUE(formulaStack.empty());
    ASSERT_TRUE(activeFormulae.empty());
    ASSERT_FALSE(analyser.getFormulae().empty());
    ASSERT_EQ(analyser.getFormulae()[0].getId(),ID);

    formula reference = formula(ID);
    reference.setOperation(ATOM);
    reference.setContent(name);

    ASSERT_EQ(analyser.getFormulae()[0].getContent(),reference.getContent());

}